# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import,
    division,
    print_function,
    with_statement,
)
import warnings

from bson.objectid import ObjectId
from pymongo import DESCENDING, collection

from turbo import mongo_model
from turbo.mongo_model import MixinModel  # noqa E401  compatibility for turbo below 0.4.5


class BaseBaseModel(mongo_model.AbstractModel):
    """class implement almost all mongodb collection method
    """

    def __init__(self, db_name='test', _mongo_db_mapping=None):
        self.__collect, self.__gridfs = super(
            BaseBaseModel, self)._init(db_name, _mongo_db_mapping)

    def __getattr__(self, k):
        attr = getattr(self.__collect, k)
        if isinstance(attr, collection.Collection):
            raise AttributeError(
                "model object '%s' has not attribute '%s'" % (self.name, k))
        return attr

    def sub_collection(self, name):
        return self.__collect[name]

    def insert(self, doc_or_docs, **kwargs):
        """Insert method
        """
        check = kwargs.pop('check', True)
        if isinstance(doc_or_docs, dict):
            if check is True:
                doc_or_docs = self._valid_record(doc_or_docs)
            result = self.__collect.insert_one(doc_or_docs, **kwargs)
            return result.inserted_id
        else:
            if check is True:
                for d in doc_or_docs:
                    d = self._valid_record(d)
            result = self.__collect.insert_many(doc_or_docs, **kwargs)
            return result.inserted_ids

    def save(self, to_save, **kwargs):
        """save method
        """
        check = kwargs.pop('check', True)
        if check:
            self._valid_record(to_save)
        if '_id' in to_save:
            self.__collect.replace_one(
                {'_id': to_save['_id']}, to_save, **kwargs)
            return to_save['_id']
        else:
            result = self.__collect.insert_one(to_save, **kwargs)
            return result.inserted_id

    def update(self, filter_, document, multi=False, **kwargs):
        """update method
        """
        self._valide_update_document(document)
        if multi:
            return self.__collect.update_many(filter_, document, **kwargs)
        else:
            return self.__collect.update_one(filter_, document, **kwargs)

    def remove(self, filter_=None, **kwargs):
        """collection remove method
        warning:
            if you want to remove all documents,
            you must override _remove_all method to make sure
            you understand the result what you do
        """
        if isinstance(filter_, dict) and filter_ == {}:
            raise ValueError('not allowed remove all documents')

        if filter_ is None:
            raise ValueError('not allowed remove all documents')

        if kwargs.pop('multi', False) is True:
            return self.__collect.delete_many(filter_, **kwargs)
        else:
            return self.__collect.delete_one(filter_, **kwargs)

    def insert_one(self, doc_or_docs, **kwargs):
        """Insert method
        """
        check = kwargs.pop('check', True)
        if check is True:
            self._valid_record(doc_or_docs)

        return self.__collect.insert_one(doc_or_docs, **kwargs)

    def insert_many(self, doc_or_docs, **kwargs):
        """Insert method
        """
        check = kwargs.pop('check', True)
        if check is True:
            for i in doc_or_docs:
                i = self._valid_record(i)

        return self.__collect.insert_many(doc_or_docs, **kwargs)

    def find_one(self, filter_=None, *args, **kwargs):
        """find_one method
        """
        wrapper = kwargs.pop('wrapper', False)
        if wrapper is True:
            return self._wrapper_find_one(filter_, *args, **kwargs)

        return self.__collect.find_one(filter_, *args, **kwargs)

    def find(self, *args, **kwargs):
        """collection find method

        """
        wrapper = kwargs.pop('wrapper', False)
        if wrapper is True:
            return self._wrapper_find(*args, **kwargs)

        return self.__collect.find(*args, **kwargs)

    @mongo_model.convert_to_record
    def _wrapper_find_one(self, filter_=None, *args, **kwargs):
        """Convert record to a dict that has no key error
        """
        return self.__collect.find_one(filter_, *args, **kwargs)

    @mongo_model.convert_to_record
    def _wrapper_find(self, *args, **kwargs):
        """Convert record to a dict that has no key error
        """
        return self.__collect.find(*args, **kwargs)

    def update_one(self, filter_, document, **kwargs):
        """update method
        """
        self._valide_update_document(document)
        return self.__collect.update_one(filter_, document, **kwargs)

    def update_many(self, filter_, document, **kwargs):
        self._valide_update_document(document)
        return self.__collect.update_many(filter_, document, **kwargs)

    def delete_many(self, filter_):
        if isinstance(filter_, dict) and filter_ == {}:
            raise ValueError('not allowed remove all documents')

        if filter_ is None:
            raise ValueError('not allowed remove all documents')

        return self.__collect.delete_many(filter_)

    def find_by_id(self, _id, projection=None):
        """find record by _id
        """
        if isinstance(_id, list) or isinstance(_id, tuple):
            return list(self.__collect.find(
                {'_id': {'$in': [self._to_primary_key(i) for i in _id]}}, projection))

        document_id = self._to_primary_key(_id)

        if document_id is None:
            return None

        return self.__collect.find_one({'_id': document_id}, projection)

    def remove_by_id(self, _id):
        if isinstance(_id, list) or isinstance(_id, tuple):
            return self.__collect.delete_many(
                {'_id': {'$in': [self._to_primary_key(i) for i in _id]}})

        return self.__collect.remove({'_id': self._to_primary_key(_id)})

    def find_new_one(self, *args, **kwargs):
        cur = list(self.__collect.find(
            *args, **kwargs).limit(1).sort('_id', DESCENDING))
        if cur:
            return cur[0]

        return None

    def get_as_dict(self, condition=None, projection=None, skip=0, limit=0, sort=None):
        as_list = self.__collect.find(
            condition, projection, skip=skip, limit=limit, sort=sort)

        as_dict, as_list = {}, []
        for i in as_list:
            as_dict[i['_id']] = i
            as_list.append(i)

        return as_dict, as_list

    def inc(self, filter_, key, num=1, multi=False):
        if multi:
            self.__collect.update_many(filter_, {'$inc': {key: num}})
        else:
            self.__collect.update_one(filter_, {'$inc': {key: num}})

    def put(self, value, **kwargs):
        if value:
            return self.__gridfs.put(value, **kwargs)
        return None

    def delete(self, _id):
        return self.__gridfs.delete(self.to_objectid(_id))

    def get(self, _id):
        return self.__gridfs.get(self.to_objectid(_id))

    def read(self, _id):
        return self.__gridfs.get(self.to_objectid(_id)).read()

    def create(self, *args, **kwargs):
        warnings.warn("create is deprecated. Use insert or insert_one "
                      "instead", DeprecationWarning, stacklevel=2)
        return self.insert(*args, **kwargs)


class BaseModel(BaseBaseModel):
    """Business model
    """

    @classmethod
    def create_model(cls, name, field=None):
        """dynamic create new model
        :args field table field, if field is None or {}, this model can not use create method
        """
        if field:
            attrs = {'name': name, 'field': field}
        else:
            attrs = {'name': name, 'field': {'_id': ObjectId()}}

        return type(str(name), (cls, ), attrs)()



# -*- coding:utf-8 -*-

from turbo.flux import Mutation, dispatch, register, register_dispatch

from . import mutation_types


@register_dispatch('user', mutation_types.INCREASE)
def increase(rank):
    pass


def decrease(rank):
    return dispatch('user', mutation_types.DECREASE, rank)



#  -*- coding:utf-8 -*-

import time
from datetime import datetime

import turbo.model
from bson.objectid import ObjectId
from pymongo import ASCENDING, DESCENDING

from .settings import MONGO_DB_MAPPING as _MONGO_DB_MAPPING

_PACKAGE_SPACE = globals()


class BaseModel(turbo.model.BaseModel):

    # for import_model work
    package_space = _PACKAGE_SPACE

    def __init__(self, db_name='test'):
        super(BaseModel, self).__init__(db_name, _MONGO_DB_MAPPING)



# -*- coding:utf-8 -*-

import turbo.log

from .base import BaseHandler
from store import actions

logger = turbo.log.getLogger(__file__)


class HomeHandler(BaseHandler):

    def get(self):
        self.render('index.html')


class IncHandler(BaseHandler):

    _get_params = {
        'option': [
            ('value', int, 0)
        ]
    }

    def get(self):
        self._data = actions.increase(self._params['value'])
        self.write(str(self._data))


class MinusHandler(BaseHandler):

    _get_params = {
        'option': [
            ('value', int, 0)
        ]
    }

    def get(self):
        self._data = actions.decrease(self._params['value'])
        self.write(str(self._data))



# -*- coding:utf-8 -*-

import tornado.options
import turbo.app
import turbo.register
from tornado.options import define, options

import setting

# uncomment this to init state manager: store
# import store

turbo.register.register_app(
    setting.SERVER_NAME,
    setting.TURBO_APP_SETTING,
    setting.WEB_APPLICATION_SETTING,
    __file__,
    globals()
)

define("port", default=8888, type=int)

if __name__ == '__main__':
    tornado.options.parse_command_line()
    turbo.app.start(options.port)



# -*- coding:utf-8 -*-


REPL_SET_NAME = 'rs0'

HOSTS = ['server1:27017', 'server2:27017', 'server3:27017']

DB_SETTING = {
    'username': 'root',
    'password': '',
    'host': 'localhost',
    'port': 3306,
    'database': 'blog'
}

DRIVERNAME = 'mysql+pymysql'


# -*- coding:utf-8 -*-
from turbo import register

from . import api
from . import app

register.register_group_urls('', [
    ('/', app.HomeHandler),
    ('/plus', app.IncHandler),
    ('/minus', app.MinusHandler),
])

register.register_group_urls('/v1', [
    ('', api.HomeHandler),
])



# -*- coding:utf-8 -*-

import os

import gridfs
from pymongo import MongoClient, MongoReplicaSetClient, read_preferences

from . import setting

mc = MongoClient(host='localhost')

# test
test = mc['test']
test_files = gridfs.GridFS(mc['test_files'])

# user
user = mc['user']
user_files = gridfs.GridFS(mc['user_files'])



from __future__ import absolute_import, division, print_function, with_statement

import multiprocessing
import os
import shutil
import signal
import socket
import tempfile
import time

import requests
from turbo import app
from turbo import register
from turbo.conf import app_config
from turbo.template import turbo_jinja2
from util import unittest


sess = requests.session()
sess.keep_alive = False

app_config.app_name = 'app_test'
app_config.app_setting['template'] = 'jinja2'
tmp_source = tempfile.mkdtemp()
TEMPLATE_PATH = os.path.join(tmp_source, "templates")
app_config.web_application_setting = {
    'xsrf_cookies': False,
    'cookie_secret': 'adasfd',
    'template_path': TEMPLATE_PATH,
    'debug': True,
}


class HomeHandler(app.BaseHandler):

    def get(self):
        self.render('index.html')

    @turbo_jinja2
    def render_string(self, *args, **kwargs):
        pass


PID = None
URL = None


def run_server(port):
    register.register_url('/', HomeHandler)
    register.register_url('', HomeHandler)
    app.start(port)


def is_used(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if sock.connect_ex(('localhost', port)) == 0:
        return True

    return False


def setUpModule():
    port = 10000
    while True:
        if not is_used(port):
            break
        port += 1

    server = multiprocessing.Process(target=run_server, args=(port,))
    server.start()
    global PID, URL
    URL = 'http://localhost:%s' % port
    PID = server.pid


def tearDownModule():
    os.kill(PID, signal.SIGKILL)


HTML = """{{'abcdefg' | upper}}"""


class AppTest(unittest.TestCase):

    def setUp(self):
        global URL
        self.home_url = URL
        try:
            shutil.rmtree(TEMPLATE_PATH)
        except:
            pass
        os.makedirs(TEMPLATE_PATH)
        with open(os.path.join(TEMPLATE_PATH, 'index.html'), 'w') as f:
            f.write(HTML)

    def test_get(self):
        time.sleep(3)
        resp = sess.get(self.home_url)
        self.assertEqual(resp.text.strip().isupper(), True)

    def tearDown(self):
        try:
            shutil.rmtree(TEMPLATE_PATH)
        except:
            pass
        finally:
            os.removedirs(tmp_source)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-

import inspect
import os
import sys
from datetime import datetime, date
import time
import json
from collections import Iterable
import copy
import base64

from bson.objectid import ObjectId

from turbo.log import util_log

PY3 = sys.version_info >= (3,)

if PY3:
    unicode_type = str
    basestring_type = str
    file_types = 'file'
    from base64 import decodebytes, encodebytes
else:
    # The names unicode and basestring don't exist in py3 so silence flake8.
    unicode_type = unicode  # noqa
    basestring_type = basestring  # noqa
    file_types = file
    from base64 import encodestring as encodebytes, decodestring as decodebytes

def to_list_str(value, encode=None):
    """recursively convert list content into string

    :arg list value: The list that need to be converted.
    :arg function encode: Function used to encode object.
    """
    result = []
    for index, v in enumerate(value):
        if isinstance(v, dict):
            result.append(to_dict_str(v, encode))
            continue

        if isinstance(v, list):
            result.append(to_list_str(v, encode))
            continue

        if encode:
            result.append(encode(v))
        else:
            result.append(default_encode(v))

    return result


def to_dict_str(origin_value, encode=None):
    """recursively convert dict content into string
    """
    value = copy.deepcopy(origin_value)
    for k, v in value.items():
        if isinstance(v, dict):
            value[k] = to_dict_str(v, encode)
            continue

        if isinstance(v, list):
            value[k] = to_list_str(v, encode)
            continue

        if encode:
            value[k] = encode(v)
        else:
            value[k] = default_encode(v)

    return value


def default_encode(v):
    """convert ObjectId, datetime, date into string
    """
    if isinstance(v, ObjectId):
        return unicode_type(v)

    if isinstance(v, datetime):
        return format_time(v)

    if isinstance(v, date):
        return format_time(v)

    return v


def to_str(v, encode=None):
    """convert any list, dict, iterable and primitives object to string
    """
    if isinstance(v, basestring_type):
        return v

    if isinstance(v, dict):
        return to_dict_str(v, encode)

    if isinstance(v, Iterable):
        return to_list_str(v, encode)

    if encode:
        return encode(v)
    else:
        return default_encode(v)


def format_time(dt):
    """datetime format
    """
    return time.mktime(dt.timetuple())


def to_objectid(objid):
    """字符对象转换成objectid
    """
    if objid is None:
        return objid

    try:
        objid = ObjectId(objid)
    except:
        util_log.error('%s is invalid objectid' % objid)
        return None

    return objid


def json_encode(data, **kwargs):
    try:
        return json.dumps(data, **kwargs)
    except:
        util_log.error('Uncaught exception in json_encode', exc_info=True)


def json_decode(data, **kwargs):
    try:
        return json.loads(data, **kwargs)
    except:
        util_log.error('Uncaught exception in json_decode', exc_info=True)


def to_int(value, default=None):
    try:
        return int(value)
    except ValueError as e:
        util_log.error(e)


def to_float(value, default=None):
    try:
        return float(value)
    except ValueError as e:
        util_log.error(e)


def to_datetime(t, micro=False):
    if micro:
        return datetime.fromtimestamp(t / 1000)
    else:
        return datetime.fromtimestamp(t)


def to_time(t, micro=False):
    if micro:
        return time.mktime(t.timetuple()) * 1000
    else:
        return time.mktime(t.timetuple())


class Escape(object):

    __slots__ = ['to_list_str', 'to_dict_str', 'default_encode', 'format_time', 'to_objectid',
                 'to_str', 'to_time', 'to_datetime', 'to_int', 'to_float', 'json_decode', 'json_encode', '__gl']

    def __init__(self, module):
        self.__gl = module

    def __getattr__(self, name):
        if name in self.__slots__:
            return self.__gl.get(name)

        raise AttributeError('escape has no attribute %s' % name)


escape = Escape(globals())


def get_base_dir(currfile, dir_level_num=3):
    """
    find certain path according to currfile
    """
    root_path = os.path.abspath(currfile)
    for i in range(0, dir_level_num):
        root_path = os.path.dirname(root_path)

    return root_path


def join_sys_path(currfile, dir_level_num=3):
    """
    find certain path then load into sys path
    """
    if os.path.isdir(currfile):
        root_path = currfile
    else:
        root_path = get_base_dir(currfile, dir_level_num)

    sys.path.append(root_path)


def import_object(name, package_space=None):
    if name.count('.') == 0:
        return __import__(name, package_space, None)

    parts = name.split('.')
    obj = __import__('.'.join(parts[:-1]),
                     package_space, None, [str(parts[-1])], 0)
    try:
        return getattr(obj, parts[-1])
    except AttributeError:
        raise ImportError("No module named %s" % parts[-1])


def camel_to_underscore(name):
    """
    convert CamelCase style to under_score_case
    """
    as_list = []
    length = len(name)
    for index, i in enumerate(name):
        if index != 0 and index != length - 1 and i.isupper():
            as_list.append('_%s' % i.lower())
        else:
            as_list.append(i.lower())

    return ''.join(as_list)


def remove_folder(path, foldername):
    if not foldername:
        return

    if not os.path.isdir(path):
        return

    dir_content = os.listdir(path)
    if not dir_content:
        return

    for item in dir_content:
        child_path = os.path.join(path, item)

        if not os.path.isdir(child_path):
            continue

        if item != foldername:
            remove_folder(child_path, foldername)
            continue

        # os.rmdir can't be allowed to deldte not empty
        for root, dirs, files in os.walk(child_path, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))

            for name in dirs:
                os.rmdir(os.path.join(root, name))
        try:
            os.rmdir(child_path)
        except Exception as e:
            raise e


def remove_file(path, filename):
    if not filename:
        return

    if not os.path.isdir(path):
        return

    dir_content = os.listdir(path)
    if not dir_content:
        return

    for item in dir_content:
        child_path = os.path.join(path, item)

        if os.path.isdir(child_path):
            remove_file(child_path, filename)
            continue

        if item != filename:
            continue

        try:
            os.remove(child_path)
        except Exception as e:
            raise e


def remove_extension(path, extension):
    if not extension:
        return

    if not os.path.isdir(path):
        return

    dir_content = os.listdir(path)
    if not dir_content:
        return

    for item in dir_content:
        child_path = os.path.join(path, item)

        if os.path.isdir(child_path):
            remove_extension(child_path, extension)
            continue

        name, ext = os.path.splitext(item)

        if ext != extension:
            continue

        try:
            os.remove(child_path)
        except Exception as e:
            raise e


def build_index(model_list):
    from turbo.model import BaseModel
    for m in model_list:
        for attr_name in dir(m):
            attr = getattr(m, attr_name)
            if inspect.isclass(attr) and issubclass(attr, BaseModel) and hasattr(attr, 'name'):
                if hasattr(attr, 'index'):
                    for index in attr.index:
                        attr().create_index(index, background=True)
                else:
                    print("model %s has no 'index' attribute" % attr.__name__)


_UTF8_TYPES = (bytes, type(None))


def utf8(value):
    # type: (typing.Union[bytes,unicode_type,None])->typing.Union[bytes,None]
    """Converts a string argument to a byte string.

    If the argument is already a byte string or None, it is returned unchanged.
    Otherwise it must be a unicode string and is encoded as utf8.
    """
    if isinstance(value, _UTF8_TYPES):
        return value
    if not isinstance(value, unicode_type):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.encode("utf-8")


_BASESTRING_TYPES = (basestring_type, type(None))


def to_basestring(value):
    """Converts a string argument to a subclass of basestring.

    In python2, byte and unicode strings are mostly interchangeable,
    so functions that deal with a user-supplied argument in combination
    with ascii string constants can use either and should return the type
    the user supplied.  In python3, the two types are not interchangeable,
    so this method is needed to convert byte strings to unicode.
    """
    if isinstance(value, _BASESTRING_TYPES):
        return value
    if not isinstance(value, bytes):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.decode("utf-8")


def get_func_name(func):
    name = getattr(func, 'func_name', None)
    if not name:
        name = getattr(func, '__name__', None)
    return name




from __future__ import absolute_import, division, print_function, with_statement

import multiprocessing
import os
import signal

from bson.objectid import ObjectId
import requests
from turbo import app
from turbo import register
from turbo.conf import app_config
from turbo.util import (
    basestring_type as basestring,
    file_types
)

from util import unittest, port_is_used, get_free_tcp_port

app_config.app_name = 'app_test'
app_config.web_application_setting = {
    'xsrf_cookies': False,
    'cookie_secret': 'adasfd'
}


class HomeHandler(app.BaseHandler):

    def get(self):
        assert not self.is_ajax()
        self.write('get')

    def post(self):
        self.write('post')

    def put(self):
        self.write('put')


class ApiHandler(app.BaseHandler):
    _get_required_params = [
        ('skip', int, 0),
        ('limit', int, 20),
    ]

    _get_params = {
        'need': [
            ('action', None),
        ],
        'option': [
            ('who', basestring, 'python'),
            ('str_key', str, 'python'),
            ('bool', bool, False),
            ('int', int, 0),
            ('float', float, 0),
            ('objectid', ObjectId, None),
            ('list', list, []),
        ]
    }

    _post_required_params = [
        ('skip', int, 0),
        ('limit', int, 20),
    ]

    _post_params = {
        'need': [
            ('who', basestring),
            ('app_test', file_types)
        ],
    }

    def GET(self):
        self._params = self.parameter

        assert self._params['skip'] == 0
        assert self._params['limit'] == 20
        assert self._params['who'] == 'python'
        assert self._params['action'] is None
        assert self._params['list'] == []
        assert self._params['str_key'] == 'python'
        assert self.is_ajax()

        self._data = {
            'value': self._params['who']
        }

    def POST(self):
        self._params = self.parameter

        assert self._params['skip'] == 0
        assert self._params['limit'] == 10

        self._data = {
            'value': self._params['who'],
            'file_name': self._params['app_test'][0].filename
        }

    def PUT(self):
        self._data = {
            'api': {
                'put': 'value'
            }
        }

    def DELETE(self):
        raise Exception('value error')

    def wo_json(self, data):
        self.write(self.json_encode(data, indent=4))


PID = None
URL = None


def run_server(port):
    register.register_url('/', HomeHandler)
    register.register_url('', HomeHandler)
    register.register_url('/api', ApiHandler)
    app.start(port)


def setUpModule():
    port = get_free_tcp_port()
    server = multiprocessing.Process(target=run_server, args=(port,))
    server.start()
    global PID, URL
    URL = 'http://localhost:%s' % port
    PID = server.pid


def tearDownModule():
    os.kill(PID, signal.SIGKILL)


class AppTest(unittest.TestCase):

    def setUp(self):
        global URL
        self.home_url = URL
        self.api_url = URL + '/api'

    def test_get(self):
        resp = requests.get(self.home_url)
        self.assertEqual(resp.status_code, 200)

    def test_post(self):
        resp = requests.post(self.home_url)
        self.assertEqual(resp.status_code, 200)

    def test_get_api(self):
        resp = requests.get(self.api_url, headers={
            'X-Requested-With': 'XMLHttpRequest'})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['res']['value'], 'python')

    def test_delete_api(self):
        resp = requests.delete(self.api_url, headers={
            'X-Requested-With': 'XMLHttpRequest'})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['msg'], 'Unknown Error')

    def test_post_api(self):
        with open(os.path.abspath(__file__), 'rb') as file:
            resp = requests.post(self.api_url, headers={
                'X-Requested-With': 'XMLHttpRequest'},
                                 data={'limit': 10, 'who': 'ruby'}, files={'app_test': file})
            self.assertEqual(resp.status_code, 200)
            self.assertEqual(resp.json()['res']['value'], 'ruby')
            self.assertEqual(resp.json()['res']['file_name'], 'app_test.py')

    def test_404(self):
        resp = requests.get(self.home_url + '/hello')
        self.assertTrue(resp.content.find(b'404') != -1)

    def test_context(self):
        """TODO test context
        """
        pass


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, division, print_function, with_statement

from turbo.util import basestring_type, unicode_type, PY3, to_basestring

if PY3:
    from urllib.parse import quote
else:
    from urllib import quote


def is_empty(v):
    if isinstance(v, basestring_type):
        if not v:
            return True
    if v is None:
        return True

    return False


def utf8(v):
    return v.encode('utf-8') if isinstance(v, unicode_type) else str(v)


def encode_http_params(**kw):
    '''
    url paremeter encode
    '''
    try:
        _fo = lambda k, v: '{name}={value}'.format(
            name=k, value=to_basestring(quote(v)))
    except:
        _fo = lambda k, v: '%s=%s' % (k, to_basestring(quote(v)))

    _en = utf8

    return '&'.join([_fo(k, _en(v)) for k, v in kw.items() if not is_empty(v)])



# -*- coding:utf-8 -*-
from __future__ import absolute_import, print_function

import re
import functools

from turbo.conf import app_config

try:
    from jinja2 import Environment, FileSystemLoader
except ImportError:
    raise ImportError('jinja2 module ImportError')


class Jinja2Environment(Environment):
    """find template location

    according to current parent and template relative path to find template path

    args:
        template current template that needs to locate
        parent which call template with extends or include directive

    return:
        real template path

    example:
        input:
            template ../../base.html
            parent app/app/index.html
        output:
            base.html

        input:
            template header.html
            parent app/app/index.html
        output:
            app/app/header.html

        input:
            template ../header.html
            parent app/app/index.html
        output:
            app/header.html

    """

    relative_path = re.compile('(./|../)', re.IGNORECASE)
    relative_dir = re.compile('([^/\s]{1,}/)', re.IGNORECASE)
    real_name = re.compile('([^/\s]{1,}$)')

    def join_path(self, template, parent):
        t_group = self.relative_path.findall(template)
        p_group = self.relative_dir.findall(parent)

        t_group_length = len(t_group)
        template_name = template
        #
        real_template_path = p_group
        if t_group_length:
            template_name = self.real_name.match(
                template, template.rfind('/') + 1).group()
            real_template_path = p_group[0:0 - t_group_length]

        real_template_path.append(template_name)
        return ''.join(real_template_path)


def turbo_jinja2(func):
    _jinja2_env = Jinja2Environment(loader=FileSystemLoader(app_config.web_application_setting[
                                    'template_path']), auto_reload=app_config.web_application_setting['debug'])

    @functools.wraps(func)
    def wrapper(self, template_name, **kwargs):
        template = _jinja2_env.get_template(
            ('%s%s') % (self.template_path, template_name))
        return template.render(handler=self, request=self.request, xsrf_form_html=self.xsrf_form_html(),
                               context=self.get_context(), **kwargs)

    return wrapper



# -*- coding:utf-8 -*-

import os
import sys

from tornado.util import ObjectDict

# server name
SERVER_NAME = '{{server_name}}'

# server dir
SERVER_DIR = os.path.dirname(os.path.abspath(__file__))
# project dir
PROJECT_DIR = os.path.dirname(SERVER_DIR)
sys.path.append(PROJECT_DIR)

# tornado web application settings
# details in
# http://www.tornadoweb.org/en/stable/web.html#tornado.web.Application.settings
WEB_APPLICATION_SETTING = ObjectDict(
    static_path=os.path.join(SERVER_DIR, "static"),
    template_path=os.path.join(SERVER_DIR, "templates"),
    xsrf_cookies=True,
    cookie_secret="3%$334ma?asdf2987^%23&^%$2",
)

# turbo app setting
TURBO_APP_SETTING = ObjectDict(
    log=ObjectDict(
        log_path=os.path.join("/var/log/", SERVER_NAME + '.log'),
        log_size=500 * 1024 * 1024,
        log_count=3,
    ),
    session_config=ObjectDict({
        'name': 'session-id',
        'secret_key': 'o387xn4ma?adfasdfa83284&^%$2'
    }),
    template='{{template_name}}',
)

# check if app start in debug
if os.path.exists(os.path.join(SERVER_DIR, '__test__')):
    WEB_APPLICATION_SETTING['debug'] = True
    TURBO_APP_SETTING.log.log_path = os.path.join("", SERVER_NAME + '.log')



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, with_statement

import datetime
import json

from bson.objectid import ObjectId
import gridfs
from pymongo import MongoClient
from turbo.model import BaseModel
from turbo.util import PY3, basestring_type as basestring, utf8
from util import unittest, fake_ids, fake_ids_2

if PY3:
    from io import StringIO
else:
    from cStringIO import StringIO

mc = MongoClient()


class Tag(BaseModel):

    name = 'tag'
    field = {
        'list': (list, []),
        'imgid': (ObjectId, None),
        'uid': (ObjectId, None),
        'name': (basestring, None),
        'value': (int, 0),
        'atime': (datetime.datetime, None),
        'up': (dict, {}),
    }

    def __init__(self):
        db = {
            'db': {'test': mc['test']},
            'db_file': {'test': gridfs.GridFS(mc['test_files'])}
        }
        super(Tag, self).__init__('test', db)

    def write_action_call(self, name, *args, **kwargs):
        pass


class BaseModelTest(unittest.TestCase):

    def setUp(self):
        self.tb_tag = Tag()
        self._make_data()

    def tearDown(self):
        self._clear_data()
        del self.tb_tag

    def _make_data(self):
        for index, i in enumerate(fake_ids):
            self.tb_tag.insert_one({'_id': i, 'value': index})

    def _clear_data(self):
        for index, i in enumerate(fake_ids):
            self.tb_tag.remove_by_id(i)

        for index, i in enumerate(fake_ids_2):
            self.tb_tag.remove_by_id(i)

    def test_insert(self):
        _id = self.tb_tag.insert({'value': 0})
        self.assertIsNot(_id, None)
        record = {
            'list': [
                {'key': ObjectId(), 'key2': 'test', 'key3': ObjectId()},
                10,
                12,
                13,
                ['name', 'name', 'name', ObjectId(), ObjectId()],
                datetime.datetime.now(),
            ],
            'imgid': ObjectId(),
            'up': {
                'key1': ObjectId(),
                'key2': ObjectId(),
                'key3': ObjectId(),
            }
        }

        result = self.tb_tag.insert(record)
        self.assertIsInstance(result, ObjectId)

        # insert one
        _id = self.tb_tag.insert({'_id': fake_ids_2[0]})
        self.assertEqual(_id, fake_ids_2[0])
        result = self.tb_tag.find_by_id(fake_ids_2[0])
        self.assertEqual(result['value'], 0)

        # insert not field
        with self.assertRaises(Exception):
            result = self.tb_tag.insert({'nokey': 10})

        # check
        _id = self.tb_tag.insert({'imgid': None})
        self.assertIsNot(_id, None)
        result = self.tb_tag.find_by_id(_id)
        self.assertEqual(result['value'], 0)
        self.tb_tag.remove_by_id(_id)

        docs = [{'_id': i} for i in fake_ids_2[1:]]
        result = self.tb_tag.insert(docs)
        self.assertEqual(1, len(result))
        for i in result:
            self.assertIn(i, fake_ids_2)
        result = self.tb_tag.find_by_id(fake_ids_2[1:])
        self.assertEqual(1, len(result))
        for i in result:
            self.assertEqual(i['value'], 0)

    def test_save(self):
        _id = self.tb_tag.save({'value': 0})
        self.assertIsNot(_id, None)
        _id = self.tb_tag.save({'_id': _id, 'value': 10})
        result = self.tb_tag.find_by_id(_id)
        self.assertEqual(result['value'], 10)
        self.tb_tag.remove_by_id(_id)

    def test_update(self):
        with self.assertRaises(ValueError):
            self.tb_tag.update({}, {'hello': 0})

        self.tb_tag.update({}, {'$set': {'hello': 0}})

        # update all empty
        with self.assertRaises(ValueError):
            self.tb_tag.update({}, {})

        # update one
        self.tb_tag.update({'_id': fake_ids[0]}, {'$set': {'value': 11}})
        result = self.tb_tag.find_by_id(fake_ids[0])
        self.assertEqual(result['value'], 11)

        # update one
        result = self.tb_tag.update({'_id': {'$in': fake_ids[1:11]}}, {
                                    '$set': {'value': 11}})
        self.assertEqual(result.matched_count, 1)
        if result.modified_count is not None:
            self.assertEqual(result.modified_count, 1)

        # update many
        result = self.tb_tag.update({'_id': {'$in': fake_ids[2:12]}}, {
                                    '$set': {'value': -1}}, multi=True)
        print(result)
        self.assertEqual(result.matched_count, 10)
        # modified_count 反应的是实际修改的文档数目，如果要修改键的值和修改值是相同的，文档不会被修改
        if result.modified_count is not None:
            self.assertEqual(result.modified_count, 10)

    def test_remove(self):
        # not allow remove all
        with self.assertRaises(Exception):
            self.tb_tag.remove({})

        result = self.tb_tag.find_by_id(fake_ids[0])
        self.assertEqual(result['_id'], fake_ids[0])

        self.tb_tag.remove({'_id': fake_ids[0]})
        result = self.tb_tag.find_by_id(fake_ids[0])
        self.assertEqual(result, None)

        # remove one
        self.tb_tag.remove({'_id': {'$in': fake_ids[2:12]}})
        result = self.tb_tag.find_by_id(fake_ids[2:12])
        self.assertEqual(len(list(result)), 9)
        # remove many
        self.tb_tag.remove({'_id': {'$in': fake_ids[3:13]}}, multi=True)
        result = self.tb_tag.find_by_id(fake_ids[3:13])
        self.assertEqual(len(list(result)), 0)

    def test_insert_one(self):
        result = self.tb_tag.insert_one({'value': 0})
        self.assertIsNot(result.inserted_id, None)

    def test_find_one(self):
        _id = self.tb_tag.insert({'value': 100})
        self.assertEqual(self.tb_tag.find_one({'_id': _id})['value'], 100)
        with self.assertRaises(KeyError):
            self.tb_tag.find_one({'_id': _id})['nokey']
        self.assertIsNone(
            self.tb_tag.find_one({'_id': _id}, wrapper=True)['nokey'])

        self.assertIsNone(self.tb_tag.find_one(
            {'_id': ObjectId()}, wrapper=True))

    def test_find(self):
        self.assertGreater(len(list(self.tb_tag.find())), 0)
        for i in list(self.tb_tag.find()):
            with self.assertRaises(KeyError):
                i['nokey']

        for i in list(self.tb_tag.find(wrapper=True)):
            self.assertIsNone(i['nokey'])

    def test_update_one(self):
        with self.assertRaises(ValueError):
            self.tb_tag.update_one({}, {'hello': 0})

        self.tb_tag.update_one({}, {'$set': {'hellow': 0}})

    def test_update_many(self):
        with self.assertRaises(ValueError):
            self.tb_tag.update_many({}, {'hello': 0})

        self.tb_tag.update_many({}, {'$set': {'value': 1}})
        for i in list(self.tb_tag.find()):
            self.assertEqual(i['value'], 1)

    def test_delete_many(self):
        with self.assertRaises(Exception):
            self.tb_tag.delete_many({})

        for i in fake_ids:
            self.tb_tag.delete_many({'_id': i})
            self.assertIsNone(self.tb_tag.find_by_id(i))

    def test_find_by_id(self):
        self.assertEqual(self.tb_tag.find_by_id(
            fake_ids[0])['_id'], fake_ids[0])
        for index, i in enumerate(self.tb_tag.find_by_id(fake_ids[0:10])):
            self.assertEqual(i['_id'], fake_ids[index])

    def test_remove_by_id(self):
        for i in fake_ids[0:10]:
            self.tb_tag.remove_by_id(i)
            self.assertIsNone(self.tb_tag.find_by_id(i))

        result = self.tb_tag.remove_by_id(fake_ids[10:20])
        for i in fake_ids[10:20]:
            self.assertIsNone(self.tb_tag.find_by_id(i))
        self.assertEqual(result.deleted_count, 10)

    def test_get_as_dict(self):
        as_dict, as_list = self.tb_tag.get_as_dict(
            {'_id': {'$in': fake_ids[0:10]}})
        for index, i in enumerate(as_dict.keys()):
            self.assertIn(i, fake_ids[0:10])

        for i in as_list:
            self.assertIn(i['_id'], fake_ids[0:10])

    def test_to_objectid(self):
        self.assertTrue(self.tb_tag.to_objectid(None) is None)
        self.assertEqual(self.tb_tag.to_objectid('52c8fb6f1d41c820f1124350'),
                         ObjectId('52c8fb6f1d41c820f1124350'), 'to_objectid is fail')

    def test_create_model(self):
        self.assertEqual(self.tb_tag.create_model(
            'tag').find_one() is not None, True)

    def test_pymongo_collection_method(self):
        self.assertEqual(self.tb_tag.full_name, 'test.tag')

    def test_sub_collection(self):
        self.assertEqual(self.tb_tag.sub_collection(
            'test').full_name, 'test.tag.test')

    def test_inc(self):
        for index, i in enumerate(self.tb_tag.find_by_id(fake_ids[0:3])):
            self.assertEqual(i['value'], index)

        self.tb_tag.inc(
            {'_id': {'$in': fake_ids[0:3]}}, 'value', 1, multi=True)
        for index, i in enumerate(self.tb_tag.find_by_id(fake_ids[0:3])):
            self.assertEqual(i['value'], index + 1)

        self.tb_tag.inc({'_id': fake_ids[4]}, 'value', 1)
        self.assertEqual(self.tb_tag.find_by_id(fake_ids[4])['value'], 5)

    def test_to_str(self):
        one = self.tb_tag.to_str(self.tb_tag.find(limit=10))
        self.assertTrue(isinstance(json.dumps(one), basestring))

    def test_to_one_str(self):
        one = self.tb_tag.to_one_str(self.tb_tag.find_one())
        self.assertTrue(isinstance(json.dumps(one), basestring))

    def test_write_action_call(self):
        def func(se, name, *args, **kwargs):
            self.assertEqual(name, 'save')

        def func2(se, name, *args, **kwargs):
            self.assertEqual(name, 'find')

        self.tb_tag.save({'value': 0})
        self.tb_tag.find()

    def test_default_encode(self):
        self.assertTrue(isinstance(
            self.tb_tag.default_encode(ObjectId()), basestring))
        self.assertTrue(isinstance(
            self.tb_tag.default_encode(datetime.datetime.now()), float))
        self.assertEqual(self.tb_tag.default_encode('string'), 'string')

    def test_put(self):
        value = 'hello word'
        s = StringIO()
        s.write(value)
        # put
        file_id = self.tb_tag.put(utf8(s.getvalue()))
        self.assertTrue(isinstance(file_id, ObjectId))

        # get
        one = self.tb_tag.get(file_id)
        self.assertTrue(getattr(one, 'read', False), 'test get fail')
        self.assertEqual(one.read(), utf8(value))

    def test_create(self):
        _id = self.tb_tag.create({'value': 0})
        self.assertIsNot(_id, None)
        record = {
            'list': [
                {'key': ObjectId(), 'key2': 'test', 'key3': ObjectId()},
                10,
                12,
                13,
                ['name', 'name', 'name', ObjectId(), ObjectId()],
                datetime.datetime.now(),
            ],
            'imgid': ObjectId(),
            'up': {
                'key1': ObjectId(),
                'key2': ObjectId(),
                'key3': ObjectId(),
            }
        }

        result = self.tb_tag.create(record)
        self.assertIsInstance(result, ObjectId)

        # create one
        _id = self.tb_tag.create({'_id': fake_ids_2[0]})
        self.assertEqual(_id, fake_ids_2[0])
        result = self.tb_tag.find_by_id(fake_ids_2[0])
        self.assertEqual(result['value'], 0)

        # create not field
        with self.assertRaises(Exception):
            result = self.tb_tag.create({'nokey': 10})

        # check
        _id = self.tb_tag.create({'imgid': None})
        self.assertIsNot(_id, None)
        result = self.tb_tag.find_by_id(_id)
        self.assertEqual(result['value'], 0)
        self.tb_tag.remove_by_id(_id)

        docs = [{'_id': i} for i in fake_ids_2[1:]]
        result = self.tb_tag.create(docs)
        self.assertEqual(1, len(result))
        for i in result:
            self.assertIn(i, fake_ids_2)
        result = self.tb_tag.find_by_id(fake_ids_2[1:])
        self.assertEqual(1, len(result))
        for i in result:
            self.assertEqual(i['value'], 0)

    def log(self, one):
        print(one)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import,
    division,
    print_function,
    with_statement
)
import sys
import os

sys.path.insert(
    0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from tests.util import unittest  # noqa E402


def main():
    testSuite = unittest.TestSuite()
    suite = unittest.TestLoader().discover('tests', pattern='*_test.py')
    testSuite.addTest(suite)

    return testSuite


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(main())



import os
import functools
import inspect
import weakref

from turbo.util import get_func_name

_mutation = {}


def register(state):
    def outwrapper(func):
        state.register(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper
    return outwrapper


class CallFuncAsAttr(object):

    class __CallObject(object):

        def __init__(self, func):
            self.func = func

        def __call__(self, *args, **kwargs):
            return self.func(*args, **kwargs)

    def __init__(self, file_attr):
        name = file_attr
        filepath = os.path.abspath(file_attr)
        if os.path.isfile(filepath):
            name, ext = os.path.splitext(os.path.basename(filepath))

        setattr(self, self._name, {})
        _mutation[name] = weakref.ref(self)

    @property
    def __get_func(self):
        return getattr(self, self._name)

    def register(self, func):
        if not inspect.isfunction(func):
            raise TypeError("argument expect function, now is '%s'" % func)

        name = get_func_name(func)
        lambda_name = get_func_name(lambda x: x)
        if name == lambda_name:
            raise TypeError('lambda is not allowed')

        self.__get_func[name] = self.__CallObject(func)

    def __getattr__(self, name):
        if name not in self.__get_func:
            raise AttributeError("%s object has no attribute '%s'" %
                                 (self.__class__.__name__, name))

        return self.__get_func[name]


class ObjectDict(dict):
    """Makes a dictionary behave like an object, with attribute-style access.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value


state = ObjectDict()


class State(object):

    __slots__ = ['_state']

    def __init__(self, file_attr):
        name = file_attr
        filepath = os.path.abspath(file_attr)
        if os.path.isfile(filepath):
            name, ext = os.path.splitext(os.path.basename(filepath))

        if name in state:
            raise KeyError('state %s has already existed' % name)

        self._state = ObjectDict()
        state[name] = self._state

    def __setattr__(self, name, value):
        if name in self.__slots__:
            return super(State, self).__setattr__(name, value)

        self._state[name] = value

    def __getattr__(self, name):
        if name not in self._state:
            raise AttributeError("%s object has no attribute '%s'" %
                                 (self.__class__.__name__, name))

        return self._state[name]


class Mutation(CallFuncAsAttr):

    @property
    def _name(self):
        return 'mutation_%s' % id(self)


def dispatch(name, type_name, *args, **kwargs):
    if name not in _mutation:
        raise ValueError('%s mutation module not found' % name)

    return getattr(_mutation[name](), type_name)(*args, **kwargs)


def register_dispatch(name, type_name):
    def outwrapper(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return dispatch(name, type_name, *args, **kwargs)

        return wrapper
    return outwrapper



# -*- coding:utf-8 -*-

from turbo.flux import Mutation, State, register

mutation = Mutation(__file__)
state = State(__file__)


@register(mutation)
def increase_rank(rank):
    return rank + 1


@register(mutation)
def decrease_rank(rank):
    return rank - 1



from __future__ import absolute_import, division, print_function, with_statement

from util import unittest

from turbo.flux import Mutation, register, dispatch, register_dispatch, State, state

mutation = Mutation('flux_test')


@register(mutation)
def increase_rank(rank):
    return rank + 1


@register(mutation)
def decrease_rank(rank):
    return rank - 1


@register_dispatch('flux_test', 'increase_rank')
def increase(rank):
    pass


def decrease(rank):
    return dispatch('flux_test', 'decrease_rank', rank)

tstate = State('test')
tstate.count = 0


class FluxTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_state(self):
        tstate.count += 1
        self.assertEqual(tstate.count, 1)
        self.assertEqual(state.test.count, 1)
        tstate.count += 1
        self.assertEqual(tstate.count, 2)
        self.assertEqual(state.test.count, 2)
        state.test.count += 1
        self.assertEqual(tstate.count, 3)
        self.assertEqual(state.test.count, 3)

    def test_increment(self):
        self.assertEqual(increase(10), 11)

    def test_decrease(self):
        self.assertEqual(decrease(10), 9)

if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-
from __future__ import (
    absolute_import,
    division,
    print_function,
    with_statement,
)

from collections import defaultdict
from datetime import datetime
import functools
import time

from bson.objectid import ObjectId
from turbo.log import model_log
from turbo.util import escape as _es, import_object


def _record(x):
    return defaultdict(lambda: None, x)


def convert_to_record(func):
    """Wrap mongodb record to a dict record with default value None
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        if result is not None:
            if isinstance(result, dict):
                return _record(result)
            return (_record(i) for i in result)

        return result

    return wrapper


class MixinModel(object):

    @staticmethod
    def utctimestamp(seconds=None):
        if seconds:
            return long(time.mktime(time.gmtime(seconds)))
        else:
            return long(time.mktime(time.gmtime()))

    @staticmethod
    def timestamp():
        return long(time.time())

    @staticmethod
    def datetime(dt=None):
        if dt:
            return datetime.strptime(dt, '%Y-%m-%d %H:%M')
        else:
            return datetime.now()

    @staticmethod
    def utcdatetime(dt=None):
        if dt:
            return datetime.strptime(dt, '%Y-%m-%d %H:%M')
        else:
            return datetime.utcnow()

    @classmethod
    def to_one_str(cls, value, *args, **kwargs):
        """Convert single record's values to str
        """
        if kwargs.get('wrapper'):
            return cls._wrapper_to_one_str(value)

        return _es.to_dict_str(value)

    @classmethod
    def to_str(cls, values, callback=None):
        """Convert many records's values to str
        """
        if callback and callable(callback):
            if isinstance(values, dict):
                return callback(_es.to_str(values))

            return [callback(_es.to_str(i)) for i in values]
        return _es.to_str(values)

    @staticmethod
    @convert_to_record
    def _wrapper_to_one_str(value):
        return _es.to_dict_str(value)

    @staticmethod
    def default_encode(v):
        return _es.default_encode(v)

    @staticmethod
    def json_encode(v):
        return _es.json_encode(v)

    @staticmethod
    def json_decode(v):
        return _es.json_decode(v)

    @staticmethod
    def to_objectid(objid):
        return _es.to_objectid(objid)

    @staticmethod
    def create_objectid():
        """Create new objectid
        """
        return ObjectId()

    _instance = {}

    @classmethod
    def instance(cls, name):
        """Instantiate a model class according to import path
        args:
            name: class import path like `user.User`
        return:
            model instance
        """
        if not cls._instance.get(name):
            model_name = name.split('.')
            ins_name = '.'.join(
                ['models', model_name[0], 'model', model_name[1]])
            cls._instance[name] = cls.import_model(ins_name)()

        return cls._instance[name]

    @classmethod
    def import_model(cls, ins_name):
        """Import model class in models package
        """
        try:
            package_space = getattr(cls, 'package_space')
        except AttributeError:
            raise ValueError('package_space not exist')
        else:
            return import_object(ins_name, package_space)

    @staticmethod
    def default_record():
        """Generate one default record that return empty str when key not exist
        """
        return defaultdict(lambda: '')


def collection_method_call(turbo_connect_ins, name):
    def outwrapper(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if name in turbo_connect_ins._write_operators:
                turbo_connect_ins._model_ins.write_action_call(
                    name, *args, **kwargs)

            if name in turbo_connect_ins._read_operators:
                turbo_connect_ins._model_ins.read_action_call(
                    name, *args, **kwargs)

            return func(*args, **kwargs)

        return wrapper

    return outwrapper


class MongoTurboConnect(object):

    _write_operators = frozenset([
        'insert',
        'save',
        'update',
        'find_and_modify',
        'bulk_write',
        'insert_one',
        'insert_many',
        'replace_one',
        'update_one',
        'update_many',
        'delete_one',
        'delete_many',
        'find_one_and_delete',
        'find_one_and_replace',
        'find_one_and_update',
        'create_index',
        'drop_index',
        'create_indexes',
        'drop_indexes',
        'drop',
        'remove',
        'ensure_index',
        'rename',
    ])

    _read_operators = frozenset([
        'find',
        'find_one',
        'count',
        'index_information',
    ])

    def __init__(self, model_ins, db_collect=None):
        self._model_ins = model_ins
        self._collect = db_collect

    def __getattr__(self, name):
        collection_method = getattr(self._collect, name)
        if callable(collection_method):
            return collection_method_call(self, name)(collection_method)

        return collection_method

    def __getitem__(self, name):
        """Sub-collection
        """
        return self._collect[name]


class AbstractModel(MixinModel):
    """
    name = None                            mongodb collection name
    field = None                           collection record map
    column = None                          need to query field
    index = [
        tuple([('uid', 1)])
    ]                                      query index
    """

    _operators = frozenset([
        '$set',
        '$unset',
        '$rename',
        '$currentDate',
        '$inc',
        '$max',
        '$min',
        '$mul',
        '$setOnInsert',

        '$addToSet',
        '$pop',
        '$pushAll',
        '$push',
        '$pull'])

    PRIMARY_KEY_TYPE = ObjectId

    def _init(self, db_name, _mongo_db_mapping):
        if _mongo_db_mapping is None:
            raise Exception("db mapping is invalid")
        # databases
        db = _mongo_db_mapping['db']
        # databases file
        db_file = _mongo_db_mapping['db_file']

        # databse name
        if db_name not in db or db.get(db_name, None) is None:
            raise Exception('%s is invalid databse' % db_name)

        # collection name
        if not self.name:
            raise Exception('%s is invalid collection name' % self.name)

        # collection field
        if not self.field or not isinstance(self.field, dict):
            raise Exception('%s is invalid collection field' % self.field)

        # collect as private variable
        collect = getattr(db.get(db_name, object), self.name, None)
        if collect is None:
            raise Exception('%s is invalid collection' % self.name)

        # replace pymongo collect with custome connect
        _collect = MongoTurboConnect(self, collect)

        # gridfs as private variable
        _gridfs = db_file.get(db_name, None)
        if _gridfs is None:
            model_log.info('%s is invalid gridfs' % _gridfs)

        return _collect, _gridfs

    def _to_primary_key(self, _id):
        if self.PRIMARY_KEY_TYPE is ObjectId:
            return self.to_objectid(_id)

        return _id

    def __setitem__(self, k, v):
        setattr(self, k, v)

    def __getitem__(self, k):
        return getattr(self, k)

    def __str__(self):
        if isinstance(self.field, dict):
            return str(self.field)
        return None

    def sub_collection(self, name):
        raise NotImplementedError()

    def find_by_id(self, _id, column=None):
        raise NotImplementedError()

    def remove_by_id(self, _id):
        raise NotImplementedError()

    def find_new_one(self, *args, **kwargs):
        """return latest one record sort by _id
        """
        raise NotImplementedError()

    def get_as_dict(self, condition=None, column=None, skip=0, limit=0, sort=None):
        raise NotImplementedError()

    def _valide_update_document(self, document):
        for opk in document.keys():
            if not opk.startswith('$') or opk not in self._operators:
                raise ValueError('invalid document update operator')

        if not document:
            raise ValueError('empty document update not allowed')

    def _valid_record(self, record):
        if not isinstance(record, dict):
            raise Exception('%s record is not dict' % record)

        rset = set(record.keys())
        fset = set(self.field.keys())
        rset.discard('_id')
        fset.discard('_id')
        if not (fset ^ rset) <= fset:
            raise Exception('record keys is not equal to fields keys %s' % (
                list((fset ^ rset) - fset)))

        for k, v in self.field.items():
            if k not in record:
                if v[0] is datetime and not v[1]:
                    record[k] = self.datetime()
                    continue

                if v[0] is time and not v[1]:
                    record[k] = self.timestamp()
                    continue

                record[k] = v[1]

        return record

    def inc(self, spec_or_id, key, num=1):
        raise NotImplementedError()

    def put(self, value, **kwargs):
        """gridfs put method
        """
        raise NotImplementedError()

    def delete(self, _id):
        """gridfs delete method
        """
        raise NotImplementedError()

    def get(self, _id):
        """gridfs get method
        """
        raise NotImplementedError()

    def read(self, _id):
        """gridfs read method
        """
        raise NotImplementedError()

    def write_action_call(self, name, *args, **kwargs):
        """Execute when write action occurs, note: in this method write action must be called asynchronously
        """
        pass

    def read_action_call(self, name, *args, **kwargs):
        """Execute when read action occurs, note: in this method read action must be called asynchronously
        """
        pass




import redis
from turbo import session


class SessionStore(session.RedisStore):

    conn = redis.Redis(host='localhost', port=6379, db=0)

    def get_connection(self, key):
        return self.conn


class SessionObject(session.CookieObject):

    def generate_session_id(self):
        _id = super(SessionObject, self).generate_session_id()
        return 'session:%s'%_id


# -*- coding:utf-8 -*-

from db.mongo_conn import (
    test as _test,
    user as _user,

    test_files as _test_files,
    user_files as _user_files,
)

MONGO_DB_MAPPING = {
    'db': {
        'test': _test,
        'user': _user,
    },
    'db_file': {
        'test': _test_files,
        'user': _user_files,
    }
}



# -*- coding:utf-8 -*-

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.url import URL

from .setting import DB_SETTING, DRIVERNAME


# mysql blog
blog_engine = create_engine(
    URL(DRIVERNAME, **DB_SETTING), encoding='utf8', echo=True)
DBSession = sessionmaker(bind=blog_engine)



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, with_statement

import sys
import socket

from bson.objectid import ObjectId

if sys.version_info < (2, 7):
    import unittest2 as unittest  # noqa E401
else:
    import unittest  # noqa E401


def port_is_used(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if sock.connect_ex(('localhost', port)) == 0:
        return True

    return False


def get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('', 0))
    addr, port = tcp.getsockname()
    tcp.close()
    return port


fake_ids = [
    ObjectId('586a01b6ed80083a5087c7d7'),
    ObjectId('586a01b6ed80083a5087c7d8'),
    ObjectId('586a01b6ed80083a5087c7d9'),
    ObjectId('586a01b6ed80083a5087c7da'),
    ObjectId('586a01b6ed80083a5087c7dc'),
    ObjectId('586a01b6ed80083a5087c7dd'),
    ObjectId('586a01b6ed80083a5087c7de'),
    ObjectId('586a01b6ed80083a5087c7df'),
    ObjectId('586a01b6ed80083a5087c7e0'),
    ObjectId('586a01b6ed80083a5087c7e1'),
    ObjectId('586a01b6ed80083a5087c7e2'),
    ObjectId('586a01b6ed80083a5087c7e3'),
    ObjectId('586a01b6ed80083a5087c7e5'),
    ObjectId('586a01b6ed80083a5087c7e6'),
    ObjectId('586a01b6ed80083a5087c7e7'),
    ObjectId('586a01b6ed80083a5087c7e8'),
    ObjectId('586a01b6ed80083a5087c7ea'),
    ObjectId('586a01b6ed80083a5087c7eb'),
    ObjectId('586a01b6ed80083a5087c7ec'),
    ObjectId('586a01b6ed80083a5087c7ed'),
    ObjectId('586a01b6ed80083a5087c7ee'),
    ObjectId('586a01b6ed80083a5087c7ef'),
    ObjectId('586a01b6ed80083a5087c7f0'),
    ObjectId('586a01b6ed80083a5087c7f1'),
    ObjectId('586a01b6ed80083a5087c7f3'),
    ObjectId('586a01b6ed80083a5087c7f4'),
    ObjectId('586a01b6ed80083a5087c7f5'),
    ObjectId('586a01b6ed80083a5087c7f6'),
    ObjectId('586a01b6ed80083a5087c7f8'),
    ObjectId('586a01b6ed80083a5087c7f9'),
    ObjectId('586a01b6ed80083a5087c7fa'),
    ObjectId('586a01b6ed80083a5087c7fb'),
    ObjectId('586a01b6ed80083a5087c7fc'),
    ObjectId('586a01b6ed80083a5087c7fd'),
    ObjectId('586a01b6ed80083a5087c7fe'),
    ObjectId('586a01b6ed80083a5087c7ff'),
    ObjectId('586a01b6ed80083a5087c801'),
    ObjectId('586a01b6ed80083a5087c802'),
    ObjectId('586a01b6ed80083a5087c803'),
    ObjectId('586a01b6ed80083a5087c804'),
    ObjectId('586a01b6ed80083a5087c806'),
    ObjectId('586a01b6ed80083a5087c807'),
    ObjectId('586a01b6ed80083a5087c808')
]

fake_ids_2 = [
    ObjectId('586a09f9ed80083a5087c809'),
    ObjectId('586a09f9ed80083a5087c80a'),
]



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, with_statement

import datetime
import json
from copy import deepcopy

from bson.objectid import ObjectId
from turbo.util import escape,  basestring_type

from util import unittest


class EscapeTest(unittest.TestCase):

    def test_inc(self):
        pass

    def test_to_str(self):
        data = {
            'v1': 10,
            'v2': datetime.datetime.now(),
            'v3': ObjectId(),
            'v4': 'value',
        }

        self.assertTrue(isinstance(json.dumps(escape.to_str(
            [deepcopy(data) for i in range(10)])), basestring_type))
        self.assertTrue(isinstance(json.dumps(
            escape.to_str(deepcopy(data))), basestring_type))

    def test_to_str_encode(self):
        data = {
            'v1': 10,
            'v2': datetime.datetime.now(),
            'v3': ObjectId(),
            'v4': 'value',
        }

        v = escape.to_str(data)

        self.assertTrue(isinstance(v['v1'], int))
        self.assertTrue(isinstance(v['v2'], float))
        self.assertTrue(isinstance(v['v3'], basestring_type))
        self.assertTrue(isinstance(v['v4'], basestring_type))

        def encode(v):
            return str(v)

        v = escape.to_str(data, encode)
        self.assertTrue(isinstance(v['v1'], basestring_type))
        self.assertTrue(isinstance(v['v2'], basestring_type))
        self.assertTrue(isinstance(v['v3'], basestring_type))
        self.assertTrue(isinstance(v['v4'], basestring_type))


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, with_statement

import copy
from datetime import datetime
import logging
import time

from bson.objectid import ObjectId
from turbo.util import escape as es, camel_to_underscore, basestring_type as basestring, unicode_type
from util import unittest


class EscapeTest(unittest.TestCase):

    def setUp(self):
        child = {
            'id': ObjectId(),
            'atime': datetime.now(),
            'number': 10,
            'name': 'hello world',
            'mail': None,
        }
        self.record = {
            'id': ObjectId(),
            'atime': datetime.now(),
            'number': 10,
            'name': 'hello world',
            'child': child,
            'childs': [copy.deepcopy(child) for i in range(3)],
        }

        self.values = [copy.deepcopy(self.record) for i in range(3)]

    def tearDown(self):
        del self.record
        del self.values

    def test_to_dict_str(self):
        self.check_value_type(es.to_dict_str(self.record))

    def test_default_encode(self):
        now = datetime.now()
        objid = ObjectId()
        number = 10
        self.assertEqual(es.default_encode(now), time.mktime(now.timetuple()))
        self.assertEqual(es.default_encode(objid), unicode_type(objid))
        self.assertEqual(es.default_encode(number), number)

    def test_recursive_to_str(self):
        now = datetime.now()
        objid = ObjectId()
        number = 10
        self.check_value_type(es.to_str(self.record))
        self.check_value_type(es.to_str(self.values))
        self.check_value_type(es.to_str(now))
        self.check_value_type(es.to_str(objid))
        self.check_value_type(es.to_str(number))

    def test_no_attribute(self):
        with self.assertRaises(AttributeError):
            es.create_objectid()

    def test_tobjectid(self):
        es.to_int('s')

    def test_json_encode(self):
        self.assertTrue(es.json_encode(
            es.to_str(self.values), indent=4) is not None)

    def test_json_decode(self):
        self.assertTrue(type(es.json_decode(
            es.json_encode(es.to_str(self.values), indent=4)
        )).__name__ == 'list'
        )

        # error test
        self.assertTrue(type(es.json_decode(
            es.json_encode(es.to_str(self.values), invlaid=4)
        )).__name__ == 'NoneType'
        )

    def test_to_list_str(self):
        [self.check_value_type(v) for v in es.to_list_str(self.values)]

    def test_camel_to_underscore(self):
        self.assertEqual(camel_to_underscore('HelloWorld'), 'hello_world')

    def check_value_type(self, v):
        if isinstance(v, list):
            [self.check_value_type(v1) for v1 in v]
            return

        if isinstance(v, dict):
            [self.check_value_type(v1) for k1, v1 in v.items()]
            return

        self.assertTrue(self.check_base_value_type(v))

    def log(self, msg):
        logging.info(msg)

    def check_base_value_type(self, v):
        return isinstance(v, int) or isinstance(v, float) or isinstance(v, basestring) or v is None


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-

from . import setting
from apps import base


class BaseHandler(base.BaseHandler):

    def initialize(self):
        super(BaseHandler, self).initialize()
        self.template_path = setting.TEMPLATE_PATH



# -*- coding:utf-8 -*-

from .base import *


class User(Model):

    """
    email: user account
    passwd: user account passwd
    atime: added time
    """
    index = [
        tuple([('email', 1)])
    ]
    name = 'user'

    field = {
        'email': (str, ''),
        'passwd': (str, ''),
        'atime': (datetime, None),
    }
